/*
**                    --- term.c ---
**
**  EXAMPLE CODE: Terminal emulator. Can transfer files using
**  ASCII, XMODEM, YMODEM, and YMODEM-G protocols.
**
**  See TERM.CFG for configuration parameters.
**
**  Link with ALLOCSEG, TERM_IO, MODEM_IO, DIR_IO, CRC, DOS, AMODEM,
**  XYMODEM, XYPACKET, SI, and OPCODES. Files SI.C and OPCODES.C used
**  for scripts only. See TERM makefiles and TERM.CFG.
**
**  Do NOT select YMODEM-G when using a null modem cable unless you are
**  certain that RTS & CTS are reversed ( this is usually not true ).
**
**  This example program (not the PCL4C library) is donated to
**  the Public Domain by MarshallSoft Computing, Inc. It is
**  provided as an example of the use of the PCL4C.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <malloc.h>
#include <dos.h>
#include "term.cfg"
#include "pcl4c.h"
#include "ascii.h"
#include "term.h"
#include "dir_io.h"
#include "term_io.h"
#include "dos_io.h"
#include "modem_io.h"
#include "crc.h"
#include "xymodem.h"
#include "amodem.h"
#include "term.h"
#include "allocseg.h"
#include "timing.h"
#if SCRIPTS
#include "si.h"
#endif

#define FALSE 0
#define TRUE !FALSE
#define NORMAL 0x07
#define INVERSE 0x70
#define MESSAGE_POS 48

/* local prototypes */
void MyStart(int Port,int BaudCode);
void ProcessESC(void);
void ShowMessage(int Pos,char *Msg);
void ShowMenuMessage(void);

/*** Global variables ***/

#define SIO_BUF_SIZE 2048

int Port;               /* current COM port [0..3] */
char FileSpec[15];      /* file name specification */
char Filename[15];      /* file name buffer */
char Buffer[1024];      /* block buffer */
char *BaudRate[10] =  {"300","600","1200","2400","4800","9600",
                       "19200","38400","57600","115200"};
int BaudCode;           /* baud rate code ( index into BaudRate[] ) */
char *ModelText[4] = {"Small","Compact","Medium","Large"};
char NCGchar   = NAK;
int  OneKflag  = FALSE;
int  BatchFlag = FALSE;
char Protocol  = 'X';
char TermChar  = 0x00;
int  CharPace  = 0;
int  Timeout   = 10;
int  EchoFlag  = TRUE;

/*** main program ***/

void main(int argc,char *argv[])
{int DataFlag = FALSE;
 char c;
 int i, k;
 int n, rc;
 int Delta;         /* delta port status */
 int Status;        /* port status */
 int Version;
 char temp[81];
 /* right number of parameters ? */
 if((argc<3)||(argc>4))
     {printf("Usage: 'TERM port baud' -- example 'TERM 1 9600 {script}'\n");
      exit(1);
     }
 /* get port number from command line */
 Port = atoi(argv[1]) - 1;
 if((Port<0) || (Port>15))
     {printf("Port must be 1 to 16\n");
      exit(1);
     }
 /* get baud rate from command line */
 BaudCode = BaudMatch(argv[2]);
 if(BaudCode<0)
     {printf("Cannot recognize baud rate = %s\n",argv[2]);
      exit(1);
     }

#if 0
 /*** Custom Configuration: 4 port card ***/
 SioIRQ(COM1,IRQ4);
 SioIRQ(COM2,IRQ3);
 /* use IRQ2 for COM3 & COM4 */
 SioIRQ(COM3,IRQ2);
 SioIRQ(COM4,IRQ2);
#endif

#if 0
 /*** Custom Configuration: DigiBoard PC/8 ***/
 puts("Configuring DigiBoard as COM1-COM8 (IRQ5)");
 SioDelay(18);
 /* use 0x140 for odd IRQs & 0x141 for even IRQs */
 SioPorts(8,COM1,0x140,DIGIBOARD);
 /* set DigiBoard UART addresses */
 ErrorCheck( SioUART(Port,0x100+8*Port) );
 /* set DigiBoard for IRQ5 */
 ErrorCheck( SioIRQ(Port,IRQ5) );
#endif

#if 0
 /*** Custom Configuration: BOCA 16 port dumbcard ***/
 puts("Configuring BOCA Board as COM1-COM16 (IRQ5)");
 SioDelay(18);
 /* use base port + 7 */
 SioPorts(16,COM1,0x100+7,BOCABOARD);
 /* set BOCA Board UART addresses */
 ErrorCheck( SioUART(Port,0x100+8*Port) );
 /* set BOCA Board for IRQ5 */
 ErrorCheck( SioIRQ(Port,IRQ5) );
#endif

 /********************************************************/

 /* setup receive buffer */
 ErrorCheck( SioRxBuf(Port,AllocSeg(2048),Size2K) );
 if(SioInfo('I'))
   {/* setup transmit buffer */
    ErrorCheck( SioTxBuf(Port,AllocSeg(2048),Size2K) );
   }
 /* set parms & reset (initialize) COM port */
 SioParms(Port,NoParity,OneStopBit,WordLength8);
 MyStart(Port,BaudCode);
 /* set DTR and RTS */
 SioDTR(Port,'S');
 SioRTS(Port,'S');
 /* init CRC table */
 InitCRC();
 /* initialize screen */
 Scroll(0,0,24,79,0,NORMAL);
 /* display status message */
 ShowMenuMessage();
 Position(1,0);
 /* display some info */
 printf("  -- TERM: 8/21/94 --\n\n");
 Version = SioInfo('V');
 printf("      Library: %d.%d\n",Version>>4,0x0f&Version);
 printf(" Memory Model: %s\n",ModelText[3&SioInfo('M')] );
 printf("TX Interrupts: ");
 if(SioInfo('I')) puts("YES");
 else puts("NO");
 printf("   Parameters: 8N1\n");
#if RTS_CTS_CONTROL
 SioFlow(Port,10*ONE_SECOND);
 printf(" Flow Control: YES\n");
#else
 printf(" Flow Control: NO\n");
#endif

#if 0
 /* Set FIFO level */
 printf("   16550 UART: ");
 if( SioFIFO(Port,LEVEL_14) ) printf("YES\n");
 else printf("NO\n");
#endif

 /* clear PCL4C receive buffer */
 ErrorCheck( SioRxFlush(Port) );
 printf("\n");

 /* see TERM.H for definition of AT_COMMAND_SET */
#if AT_COMMAND_SET
 /* wait for Modem to say its ready */
 printf("  <<Waiting for Modem DSR>>\n");
 while( !SioDSR(Port) )
     {if(SioKeyPress()||SioBrkKey()) MyExit(0,"Aborted by user");
      putchar('.');
      SioDelay(18);
     }
 printf("  <<DSR on>>\n");
#endif

#if RTS_CTS_CONTROL
 printf("  <<Waiting for Modem CTS>>\n");
 while( !SioCTS(Port) )
     {if(SioKeyPress()||SioBrkKey()) MyExit(0,"Aborted by user");
      putchar('.');
      SioDelay(18);
     }
 printf("  <<CTS on>>\n");
#endif

#if AT_COMMAND_SET
 /* initialize (Hayes compatible) modem */
 ModemSendTo(Port,4,"AT!");
 ModemEcho(Port,10);
 ModemSendTo(Port,4,"AT E1 S7=60 S11=60 V1 X1 Q0 S0=1!");
 if(ModemWaitFor(Port,5*ONE_SECOND,TRUE,"OK"))
   {/* modem is ready */
    puts("\n  <<MODEM READY>>");
   }
 else printf("\n  <<WARNING: Expected OK not received>>\n");
#endif

#if SCRIPTS
 if(argc==4)
   {/* call script interpreter */
    i = si(Port,argv[3],FALSE);
    if(i<0) SaySiErr(i);
   }
#endif

 /* enter terminal loop */
 SioRxFlush(Port);
 puts("  <<Entering terminal loop>>");
 while(1)
     {/* Control-BREAK ? */
      if(SioBrkKey()) MyExit(0,"User pressed Ctrl-BREAK");
      /* was key pressed ? */
      if(SioKeyPress())
          {/* read key press */
           i = SioKeyRead();

if((char)i=='#') printf("RX count = %d\n", SioInfo('R') );

           if((char)i==ESC)
               {/* process user's request */
                ProcessESC();
                ShowMenuMessage();
                continue;
               }
           else SioPutc(Port,(char)i);
          }
      /* was break detected ? */
      if( SioBrkSig(Port,'D') ) DisplayLine("BREAK detected ",NULL,0);
      /* any incoming over serial port ? */
      i = SioGetc(Port,1);
      if(i>-1)
          {/* good character */
           if(DataFlag&((i<0x20)||(i>0x7e)))
               {MyCrtWrite('^');
                MyCrtWrite((char)('@'+i));
               }
           else MyCrtWrite((char)i);
          }
     } /* end -- key pressed */
}

/*** write to screen except for bottom line ***/

void MyCrtWrite(char ch)
{/* write character */
 SioCrtWrite(ch);
 /* scroll all but bottom line */
 if(GetRow()==24)
    {Scroll(0,0,23,79,1,NORMAL);
     Position(23,0);
    }
}

/*** make multiple attempts to reset port ***/

void MyStart(int Port,int BaudCode)
{int i, rc;
 /* try up to 3 times to reset COM port */
 for(i=0;i<3;i++)
     {printf("Resetting COM%d at %s baud\n",Port+1,BaudRate[BaudCode]);
      if( (rc = SioReset(Port,BaudCode))==0) return;
      if(rc<0) MyExit(rc,"Error resetting port");
      SioDone(Port);
      /* display errors */
      if(rc&OverrunError) puts("Overrun Error");
      if(rc&ParityError)  puts("Parity Error");
      if(rc&FramingError) puts("Framing Error");
      if(rc&BreakDetect)  puts("Break Detect");
     }
 exit(1);
}

/*** find baud rate string in table ***/

int BaudMatch(char *ptr)
{int i;
 /* find baud rate in table */
 for(i=0;i<10;i++) if(strcmp(BaudRate[i],ptr)==0) return(i);
 return(-1);
}

/*** user pressed Escape */

void ProcessESC(void)
{int i;
 int rc;
 int c1, c2;
 char Answer[2]; /* array for 1 char answer */
 int row, col;
 /* user pressed <ESC> */
 Answer[0] = '?';
 Answer[1] = '\0';
 DisplayLine("Q)uit P)rotocol S)end R)eceive: ",Answer,1);
 if(strlen(Answer)) switch(toupper(Answer[0]))
     {
      case 'P':
#if RTS_CTS_CONTROL
          DisplayLine("A)scii X)modem Y)modem ymodem-G): ",Answer,1);
#else
          DisplayLine("A)scii X)modem Y)modem: ",Answer,1);
#endif
          if(strlen(Answer)) switch( toupper(Answer[0]) )
               {
                case 'A':
                    Protocol = 'A';
                    ShowMenuMessage();
                    TermChar = 0x18;  /* CAN or control-X */
                    CharPace = 0;     /* no inter-byte delay */
                    Timeout = 7;      /* timeout after 7 seconds */
                    EchoFlag = TRUE;  /* do local echo */
                    DisplayLine("Protocol = ASCII",NULL,1);
                    break;
                case 'X':
                    Protocol = 'X';
                    ShowMenuMessage();
                    OneKflag = FALSE;
                    BatchFlag = FALSE;
                    NCGchar = NAK;
                    DisplayLine("Protocol = XMODEM",NULL,1);
                    break;
                case 'Y':
                    Protocol = 'Y';
                    ShowMenuMessage();
                    OneKflag = TRUE;
                    BatchFlag = TRUE;
                    NCGchar = 'C';
                    DisplayLine("Protocol = YMODEM",NULL,1);
                    break;
#if RTS_CTS_CONTROL
                case 'G':
                    Protocol = 'G';
                    ShowMenuMessage();
                    OneKflag = TRUE;
                    BatchFlag = TRUE;
                    NCGchar = 'G';
                    DisplayLine("Protocol = YMODEM-G",NULL,1);
                    break;
#endif
                default:
                    DisplayLine("Must answer X, Y, or G",NULL,1);
                    break;
               }
          break;
      case 'Q':
          MyExit(0,"User pressed ESC");
          break;
      case 'R':
          ShowMessage(strlen(BaudRate[BaudCode])+9,"'CTRL-X aborts'");
          if(Protocol=='A')
              {/* Ascii */
               DisplayLine("Enter filename:",Filename,15);
               if(strlen(Filename)==0) break;
               RxAscii(Port,Filename,Buffer,1024,SIO_BUF_SIZE,FALSE,TermChar,Timeout,EchoFlag);
               break;
              }
          /* XMODEM / YMODEM receive */
          if(BatchFlag) YmodemRx(Port,Buffer,NCGchar);
          else XmodemRx(Port,NULL,Buffer,NCGchar);
          break;

      case 'S':
          ShowMessage(strlen(BaudRate[BaudCode])+9,"'CTRL-X aborts'");
          if(Protocol=='A')
              {/* Ascii */
               DisplayLine("Enter filename:",FileSpec,15);
               if(strlen(FileSpec)==0) break;
               TxAscii(Port,FileSpec,Buffer,1024,FALSE,CharPace,TermChar,EchoFlag);
               break;
              }
          /* XMODEM / YMODEM send */
          if(BatchFlag) YmodemTx(Port,NULL,Buffer,OneKflag);
          else XmodemTx(Port,NULL,Buffer,OneKflag);
          break;
     default:
          DisplayLine("Must answer Q, P, S, or R",NULL,0);
          break;
    } /* end switch */
}

/*** show menu message ***/

void ShowMenuMessage(void)
{char temp[80];
 sprintf(temp," COM%d %s %c 'ESC for menu'  ",1+Port,BaudRate[BaudCode],Protocol);
 ShowMessage(0,temp);
}

/*** show chosen message ***/

void ShowMessage(int Pos,char *Msg)
{int i;
 int SaveRow;
 int SaveCol;
 int Col;
 SaveRow = GetRow();
 SaveCol = GetCol();
 Col = MESSAGE_POS+Pos;
 for(i=0;i<strlen(Msg);i++)
   {Position(24,(unsigned char)Col++);
    AttrWrite(Msg[i],INVERSE);
   }
 Position((unsigned char)SaveRow,(unsigned char)SaveCol);
}

/*** exit program ***/

void MyExit(int code, char *msg)
{int rc;
 if(code<0) SioError(code);
 SioDone(Port);
 printf("\nTERMINATING: %s\n",msg);
 exit(0);
}

/*** check for error ***/

void ErrorCheck(int Code)
{/* trap PCL error codes */
 if(Code<0)
     {SioError(Code);
      SioDone(Port);
      exit(1);
     }
} /* end ErrorCheck */

int GetBufSize(void)
{int BufSize = SIO_BUF_SIZE;
 return BufSize;
}
